<?php

/**
 *  ▄▄▄·  ▄▄▄· ▄▄▄· ▄▄▄▄▄ ▄ .▄ ▄· ▄▌
 * ▐█ ▀█ ▐█ ▄█▐█ ▀█ •██  ██▪▐█▐█▪██▌
 * ▄█▀▀█  ██▀·▄█▀▀█  ▐█.▪██▀▐█▐█▌▐█▪
 * ▐█ ▪▐▌▐█▪·•▐█ ▪▐▌ ▐█▌·██▌▐▀ ▐█▀·.
 *  ▀  ▀ .▀    ▀  ▀  ▀▀▀ ▀▀▀ ·  ▀ •
 *  <https://fortreeforums.xyz/>
 */

namespace apathy\StyleSuite\Repository;

use apathy\StyleSuite\Helper;
use apathy\StyleSuite\Entity\Postbit;
use apathy\StyleSuite\Entity\Username;
use apathy\StyleSuite\Service\User\Postbit as PostbitService;
use apathy\StyleSuite\Service\User\UsernameIcon;
use XF;
use XF\Mvc\Entity\Repository;

class Styles extends Repository
{
    /** The names of the cache columns in xf_data_registry */
    protected $cacheId = [
        'avatar_frame'       => 'apAvatarFrames',
        'font'               => 'apFonts',
        'postbit_preset'     => 'apPostbitBgs',
        'sparkle'            => 'apSparkles',
        'custom_title_glow'  => 'apTitleGlows',
        'custom_title_style' => 'apTitleStyles',
        'postbit_style'      => 'apPostbits',
        'tooltip_style'      => 'apTooltips',
        'username_glow'      => 'apUnameGlows',
        'username_style'     => 'apUnameStyles'
    ];

    public function findFontsOnFileSystem(): array
    {
        $dir = $this->app()->config('externalDataPath') . '/fonts/apathy/StyleSuite/';
        $dir = strtr($dir, ['data' => 'styles']);
        $fonts = glob($dir . '*.{otf,ttf,woff,woff2,OTF,TTF,WOFF,WOFF2}', GLOB_BRACE);

        if(!$fonts || !is_array($fonts))
        {
            return [];
        }

        $fontlist = [];

        foreach($fonts as $font)
        {
            $fontlist[] = strtr($font, [$dir => '']);
        }

        return $fontlist;
    }

    public function findFramesOnFileSystem(): array
    {
        return $this->findAssetsOnFileSystem('apathy/StyleSuite/frames/');
    }
    
    public function findSparklesOnFileSystem(): array
    {
        return $this->findAssetsOnFileSystem('apathy/StyleSuite/sparkles/');
    }

    public function findPostbitPresetsOnFileSystem(): array
    {
        return $this->findAssetsOnFileSystem('apathy/StyleSuite/postbits/presets');
    }

    public function getFontsFromCache(): array
    {
        return $this->app()->registry()->get($this->cacheId['font']);
    }

    public function getFramesFromCache(): array
    {
        return $this->app()->registry()->get($this->cacheId['avatar_frame']);
    }

    public function getPostbitPresetsFromCache(): array
    {
        return $this->app()->registry()->get($this->cacheId['postbit_preset']);
    }

    public function getSparklesFromCache(): array
    {
        return $this->app()->registry()->get($this->cacheId['sparkle']);
    }

    public function rebuildFontCache()
    {
        return $this->rebuildAssetCache(
            $this->findFontsOnFileSystem(), $this->cacheId['font']
        );
    }

    public function rebuildFrameCache()
    {
        return $this->rebuildAssetCache(
            $this->findFramesOnFileSystem(), $this->cacheId['avatar_frame']
        );
    }

    public function rebuildSparkleCache()
    {
        return $this->rebuildAssetCache(
            $this->findSparklesOnFileSystem(), $this->cacheId['sparkle']
        );
    }

    public function rebuildPostbitPresetCache()
    {
        return $this->rebuildAssetCache(
            $this->findPostbitPresetsOnFileSystem(), $this->cacheId['postbit_preset']
        );
    }
    
    /**
     * @todo 
     * - When style columns have been renamed, convert this function to
     * a wrapper for rebuildUserStyleGlowCache()
     */
    public function rebuildCustomTitleStyleCache()
    {
        $users = $this->finder('XF:User')->where('ap_title_style', '!=', NULL)->fetch();

        $style = [];

        foreach($users as $user)
        {
            $ttl = $user->ap_title_style;

            $style[$user->user_id] = [
                'bold'       => $ttl['ttl_bold']       ?? NULL,
                'italic'     => $ttl['ttl_italic']     ?? NULL,
                'underline'  => $ttl['ttl_underline']  ?? NULL,
                'color'      => $ttl['ttl_color']      ?? NULL,
                'color_type' => $ttl['ttl_color_type'] ?? NULL,
                'color_end'  => $ttl['ttl_color_end']  ?? NULL,
                'font'       => $ttl['ttl_font']       ?? NULL,
                'glow'       => $ttl['ttl_glow']       ?? NULL,
                'glow_alt'   => $ttl['ttl_glow_alt']   ?? NULL,
                'glow_str'   => $ttl['ttl_glow_str']   ?? NULL,
                'bg'         => $ttl['ttl_bg']         ?? NULL
            ];
        }

        $this->app()->registry()->set($this->cacheId['custom_title_style'], $style);

        /** @var \XF\Repository\Style $styles */
        $styles = $this->repository('XF:Style');
        $styles->updateAllStylesLastModifiedDate();

        return $style;
    }
    
    public function rebuildUsernameGlowCache()
    {
        return $this->rebuildUserStyleGlowCache(
            'apathy\StyleSuite:Username', $this->cacheId['username_glow']
        );
    }
    
    /**
     * @todo 
     * - When style columns have been renamed, convert this function to
     * a wrapper for rebuildUserStyleGlowCache()
     */
    public function rebuildCustomTitleGlowCache()
    {
        $users = $this->finder('XF:User')->where('ap_title_style', '!=', NULL)->fetch();

        $animation = [];

        foreach($users as $user)
        {
            $to   = $user->ap_title_style['ttl_glow']     ?? NULL;
            $from = $user->ap_title_style['ttl_glow_alt'] ?? NULL;

            if($to !== NULL && $from !== NULL)
            {
                $animation[$user->user_id] = [
                    'from' => $to,
                    'to'   => $from
                ];
            }
        }

        $this->app()->registry()->set($this->cacheId['custom_title_glow'], $animation);

        /** @var \XF\Repository\Style $style */
        $style = $this->repository('XF:Style');
        $style->updateAllStylesLastModifiedDate();

        return $animation;
    }

    /**
     * @todo 
     * - When style columns have been renamed, convert this function to
     * a wrapper for rebuildUserStyleGlowCache()
     */
    public function rebuildTooltipStyleCache()
    {
        $users  = $this->finder('XF:User')->where('ap_tooltip_style', '!=', NULL)->fetch();

        $styles = [];

        foreach($users as $user)
        {
            $style = $user->ap_tooltip_style;

            $styles[$user->user_id] = [
                'color'     => $style['ttip_font_color'] ?? NULL,
                'urlColor'  => $style['ttip_url_color']  ?? NULL,
                'highlight' => $style['ttip_hlight']     ?? NULL,
                'boxBg'     => $style['ttip_box_bg']     ?? NULL,
                'border'    => $style['ttip_border']     ?? NULL,
                'button'    => $style['ttip_button']     ?? NULL
            ];
        }

        $this->app()->registry()->set($this->cacheId['tooltip_style'], $styles);

        /** @var \XF\Repository\Style $style */
        $style = $this->repository('XF:Style');
        $style->updateAllStylesLastModifiedDate();

        return $styles;
    }

    /**
     * @todo 
     * All style rebuild functions will follow this format
     * as they begin to be migrated to entities.
     * 
     * Completed:
     * Postbit, Username
     * 
     * Pending:
     * Title, tooltip, post
     */
    public function rebuildPostbitStyleCache()
    {
        return $this->rebuildUserStyleCache(
            'apathy\StyleSuite:Postbit', $this->cacheId['postbit_style']
        );
    }

    public function rebuildUsernameStyleCache()
    {
        return $this->rebuildUserStyleCache(
            'apathy\StyleSuite:Username', $this->cacheId['username_style']
        );
    }

    /** We only perform a soft-delete to avoid triggering file health check warnings */
    public function setDefaultItemAsDeleteUndeleted($item, $cache, $registry, $deleted = true)
    {
        $cache[$item]['deleted'] = $deleted;

        $this->app()->registry()->set($registry, $cache);

        return $cache;
    }

    public function setItemAttribute($item, string $type, $cache, $registry)
    {
        foreach($item as $key => $value)
        {
            $fallbackValue = ($type == 'name') ? $key : $value;

            $cache[$key][$type] = $value ?: $fallbackValue;
        }

        $this->app()->registry()->set($registry, $cache);

        return $cache;
    }

    public function sortItemsByAttribute($cache): array
    {
        if(empty($cache))
        {
            return [];
        }

        $data = [];

        foreach($cache as $key => $value)
        {
            if(isset($value['hidden']) && $value['hidden'] == 1)
            {
                $data['hidden'][$key] = $value;
            }

            if(isset($value['special']) && $value['special'] == 1)
            {
                $data['special'][$key] = $value;
            }

            if(isset($value['special']) && isset($value['hidden']))
            {
                if($value['hidden'] == 0 && $value['special'] == 0)
                {
                    $data['normal'][$key] = $value;
                }
            }
        }

        return $data;
    }

    public function uploadPostbitImage(PostbitService $service, Postbit $style, 
    $upload): string
    {
        if($upload)
        {
            if(!$service->setImageFromUpload($upload))
            {
                throw new XF\PrintableException($service->getError());
            }

            if(!$service->updatePostbit())
            {
                throw new XF\PrintableException(
                    XF::phrase('ap_ss_new_postbit_could_not_be_processed')
                );
            }     
            
            return $style->getPostbitUrl('l');
        }
            
        return $style->background;
    }

    public function uploadUsernameIconImage(UsernameIcon $service, Username $style, 
    $upload): string
    {
        if($upload)
        {
            if(!$service->setImageFromUpload($upload))
            {
                throw new XF\PrintableException($service->getError());
            }
        
            if(!$service->updateIcon())
            {
                throw new XF\PrintableException(
                    XF::phrase('ap_ss_new_icon_could_not_be_processed')
                );
            }
                            
            return $style->getIconUrl('l');
        }

        return $style->icon;
    }

    public function isValidAsset($asset, string $type)
    {
        $filename = $asset->getTempFile();

        if($type !== 'font')
        {
            $acceptedExtensions = ['jpeg','jpg','gif','png','webp'];
        }
        else
        {
            $acceptedExtensions = ['ttf','otf','woff','woff2'];
        }

        $type = strtr($type, '_', ' ');

        if(!file_exists($filename))
		{
			throw new XF\PrintableException(
                "Invalid file '$filename' passed to '$type' service" 
            );
		}

		if(!is_readable($filename))
		{
			throw new XF\PrintableException(
                "'$filename' passed to '$type' service is not readable"
            );
		}

		if(!in_array($asset->getExtension(), $acceptedExtensions))
		{
			throw new XF\PrintableException(
                XF::phrase('ap_ss_provided_file_is_not_a_valid_asset')
            );

			return false;
		}

        return true;
    }

    public function findGroupStyle($group_id)
    {
        $finder = $this->finder('Truonglv\Groups:Group');
        $finder = $finder->where('group_id', $group_id)->fetchOne();
        
        return $finder->ap_group_style;
    }

    protected function findAssetsOnFileSystem(string $path): array
    {
        $app = $this->app();

        $dir = $app->config('externalDataPath').'://' . $path;
        $files = $app->fs()->listContents($dir);

        if(!$files || !is_array($files))
        {
            return [];
        }

        $acceptedExtensions = [
            'jpeg', 'jpg', 'gif', 'png',
            'webp', 'ttf', 'otf', 'woff', 
            'woff2'
        ];

        $items = [];

        foreach($files as $file)
        {
            if(in_array($file['extension'], $acceptedExtensions))
            {
                $items[] = $file['basename'];
            }
        }

        return $items;
    }

    protected function rebuildAssetCache(array $items, string $cacheId)
    {
        $registry = $this->app()->registry();
        $cache = $registry->get($cacheId);

        $styles = [];

        foreach($items as $item)
        {
            $name = Helper::stripFileExtensions($item);

            $styles[$item] = [
                'hidden'  => $cache[$item]['hidden']  ?? false,
                'special' => $cache[$item]['special'] ?? false,
                'name'    => $cache[$item]['name']    ?? $name,
                'deleted' => $cache[$item]['deleted'] ?? false
            ];
        }

        $registry->set($cacheId, $styles);

        return $styles;
    }

    protected function rebuildUserStyleCache(string $finderId, string $cacheId)
    {
        $users = $this->finder($finderId)->fetch();

        $styles = [];

        foreach($users as $user)
        {
            $user = $user->toArray();

            foreach($user as $column => $row)
            {
                $styles[$user['user_id']][$column] = $row;
            }
        }

        unset($styles['style_id']);

        $this->app()->registry()->set($cacheId, $styles);

        /** @var \XF\Repository\Style $style */
        $style = $this->repository('XF:Style');
        $style->updateAllStylesLastModifiedDate();

        return $styles;
    }

    protected function rebuildUserStyleGlowCache(string $finderId, string $cacheId)
    {
        $users = $this->finder($finderId)
                      ->where('glow', '!=', '')
                      ->where('glow_alt', '!=', '')
                      ->fetch();

        $styles = [];

        foreach($users as $user)
        {
            $styles[$user->user_id] = [
                'from' => $user['glow'],
                'to'   => $user['glow_alt']
            ];
        }

        $this->app()->registry()->set($cacheId, $styles);

        /** @var \XF\Repository\Style $style */
        $style = $this->repository('XF:Style');
        $style->updateAllStylesLastModifiedDate();

        return $styles;
    }

    /**
     * @todo 
     * - Move this to eventual Tooltip entity
     * - Also come up with a better name
     * - Actually implement it
     */
    public function setDataForTooltips()
    {
        $visitor = XF::visitor();

        $attributes = [];

        if($visitor->Option->ap_ss_disable_user_styles)
        {
            $attributes['data-disable-user-styles'] = true;
        }

        if($visitor->canChangeTooltipStyle())
        {
            $attributes['data-can-change-tooltip-style'] = true;
        }

        if(!empty($visitor->UserStyleIgnored))
        {
            $attributes['data-ignored-user-styles'] = $visitor->findIgnoredUserStyles();
        }

        return $attributes;
    }
}
