<?php

/**
 *  ▄▄▄·  ▄▄▄· ▄▄▄· ▄▄▄▄▄ ▄ .▄ ▄· ▄▌
 * ▐█ ▀█ ▐█ ▄█▐█ ▀█ •██  ██▪▐█▐█▪██▌
 * ▄█▀▀█  ██▀·▄█▀▀█  ▐█.▪██▀▐█▐█▌▐█▪
 * ▐█ ▪▐▌▐█▪·•▐█ ▪▐▌ ▐█▌·██▌▐▀ ▐█▀·.
 *  ▀  ▀ .▀    ▀  ▀  ▀▀▀ ▀▀▀ ·  ▀ •
 *  <https://fortreeforums.xyz/>
 */

namespace apathy\StyleSuite\Service\User;

use apathy\StyleSuite\Entity\Postbit as Style;
use XF;
use XF\App;
use XF\Entity\User;
use XF\Http\Upload;
use XF\Service\AbstractService;
use XF\Util\File;

class Postbit extends AbstractService
{
	/**
	 * @var User
	 */
	protected $user;

	/**
	 * @var Style
	 */
	protected $style;

	protected $logIp = true;
	protected $logChange = true;

	protected $fileName;

	protected $type;

	protected $error = null;

	protected $allowedTypes = [IMAGETYPE_GIF, IMAGETYPE_JPEG, IMAGETYPE_PNG];

	protected $sizeMap;

	protected $throwErrors = true;

	public function __construct(App $app, User $user, Style $style)
	{
		parent::__construct($app);

        $options = XF::options();

		$this->setUser($user);
		$this->setStyle($style);

		$this->sizeMap = [
            'l' => $options->apSsMaxPostbitWidth,
            'm' => round( $options->apSsMaxPostbitWidth / 1.5 )
        ];
	}

	protected function setUser(User $user)
	{
		if($user->user_id)
		{
			$this->user = $user;
		}
		else
		{
			throw new \LogicException("User must be saved");
		}
	}

    /** @todo Possibly unused */
	protected function setStyle(Style $style)
	{
		$this->style = $style;
	}

	public function logIp($logIp)
	{
		$this->logIp = $logIp;
	}

	public function logChange($logChange)
	{
		$this->logChange = $logChange;
	}

	public function getError()
	{
		return $this->error;
	}

	public function silentRunning($runSilent)
	{
		$this->throwErrors = !$runSilent;
	}

	public function importImage($image)
	{
		$this->setImage($image);
	}

	public function setImage($fileName)
	{
		if(!$this->validateImageAsPostbit($fileName, $error))
		{
			$this->error = $error;
			$this->fileName = null;

			return false;
		}

		$this->fileName = $fileName;

		return true;
	}

	public function setImageFromUpload(Upload $upload)
	{
		$upload->requireImage();

		if(!$upload->isValid($errors))
		{
			$this->error = reset($errors);

			return false;
		}

		return $this->setImage($upload->getTempFile());
	}

	public function validateImageAsPostbit($fileName, &$error = null)
	{
		$error = null;

		if(!file_exists($fileName))
		{
			return $this->throwException(
                new \InvalidArgumentException(
                    "Invalid file '$fileName' passed to postbit service"
                )
            );
		}

		if(!is_readable($fileName))
		{
			return $this->throwException(
                new \InvalidArgumentException(
                    "'$fileName' passed to postbit service is not readable"
                )
            );
		}

		$imageInfo = filesize($fileName) ? @getimagesize($fileName) : false;

		if(!$imageInfo)
		{
			$error = XF::phrase('provided_file_is_not_valid_image');

			return false;
		}

		$type = $imageInfo[2];

		if(!in_array($type, $this->allowedTypes))
		{
			$error = XF::phrase('provided_file_is_not_valid_image');

			return false;
		}

		$width = $imageInfo[0];
		$height = $imageInfo[1];

		if(!$this->app->imageManager()->canResize($width, $height))
		{
			$error = XF::phrase('uploaded_image_is_too_big');

			return false;
		}

		$this->type = $type;

		return true;
	}

	public function updatePostbit()
	{
		if(!$this->fileName)
		{
			return $this->throwException(
                new \LogicException("No source file for postbit set")
            );
		}

		if(!$this->user->exists())
		{
			return $this->throwException(
                new \LogicException("User does not exist, cannot update postbit")
            );
		}

		$imageManager = $this->app->imageManager();

		$outputFiles = [];

		foreach($this->sizeMap AS $size => $width)
		{
			$image = $imageManager->imageFromFile($this->fileName);

			if(!$image)
			{
				continue;
			}

			$image->resizeWidth($width);

			$newTempFile = File::getTempFile();

			if($newTempFile && $image->save($newTempFile, null, 95))
			{
				$outputFiles[$size] = $newTempFile;
			}

			unset($image);
		}

		if(count($outputFiles) != count($this->sizeMap))
		{
			return $this->throwException(
                new \RuntimeException(
                    "Failed to save image to temporary file; image may be corrupt or check internal_data/data permissions"
                )
            );
		}

		foreach($outputFiles AS $code => $file)
		{
			$dataFile = $this->style->getAbstractedPostbitPath($code);

			File::copyFileToAbstractedPath($file, $dataFile);
		}

		$postbit = $this->style;

		$postbit->bg_date = XF::$time;

		if($this->logChange == false)
		{
			$postbit->getBehavior('XF:ChangeLoggable')->setOption('enabled', false);
		}

		if($this->logIp)
		{
			$ip = ($this->logIp === true ? $this->app->request()->getIp() : $this->logIp);

			$this->writeIpLog('update', $ip);
		}

		return true;
	}

	public function deletePostbit()
	{
		$this->deletePostbitFiles();

		$postbit = $this->user->PostbitStyle;

		$postbit->bg_date = 0;

		$postbit->save();

		if($this->logIp)
		{
			$ip = ($this->logIp === true ? $this->app->request()->getIp() : $this->logIp);

			$this->writeIpLog('delete', $ip);
		}

		return true;
	}

	public function deletePostbitForUserDelete()
	{
		$this->deletePostbitFiles();

		return true;
	}

	protected function deletePostbitFiles()
	{
		if($this->user->PostbitStyle->bg_date)
		{
			foreach($this->sizeMap AS $code => $size)
			{
				File::deleteFromAbstractedPath(
                    $this->user->PostbitStyle->getAbstractedPostbitPath($code)
                );
			}
		}
	}

	protected function writeIpLog($action, $ip)
	{
		$user = $this->user;

		/** @var \XF\Repository\Ip $ipRepo */
		$ipRepo = $this->repository('XF:Ip');
		$ipRepo->logIp(
            XF::visitor()->user_id, $ip, 'user', $user->user_id, 'postbit_bg_' . $action
        );
	}

	protected function throwException(\Exception $error)
	{
		if ($this->throwErrors)
		{
			throw $error;
		}
		else
		{
			return false;
		}
	}
}